#include <stdlib.h>
#include <limits.h>
#include "util.h"
#include "align.h"

static inline void
min_column(long *d, int from, int to, int from_column, int to_column,
           int column_length, int diagonal_width)
{
  int diff = to_column - from_column;
  int from_lb = max2((diff > 0) ? diff : 0,
                     1 - (from_column - (diagonal_width / 2)));
  int to_lb = max2((diff < 0) ? -diff : 0,
                   1 - (to_column - (diagonal_width / 2)));
  int from_ub = min2(diagonal_width,
                     column_length - (from_column - (diagonal_width / 2)));
  int to_ub = min2(diagonal_width,
                   column_length - (to_column - (diagonal_width / 2)));

  for (int f = from_lb, t = to_lb;
       f < from_ub && t < to_ub;
       f++, t++) {
    if (d[(f * 2) + from] < d[(t * 2) + to]) {
      d[(t * 2) + to] = d[(f * 2) + from];
    }
  }
}

static inline void
calculate_column(long *d, int offset,
                 int column_length,
                 int column, int diagonal_width,
                 char *A, char B_char)
{
  int lb = max2(0, 1 - (column - (diagonal_width / 2))); 
  int ub = min2(diagonal_width,
                column_length - (column - (diagonal_width / 2)));

  for (int i = lb; i < ub; i++) {
    int io = (i * 2) + offset;
    d[io] = min3(d[io + 2] + 1,
                 d[io] + (A[i + (column - (diagonal_width / 2)) - 1] != B_char),
                 d[io - 2] + 1);
  }
}

static long _align(char *A,
                   int A_length,
                   char *B1,
                   int B1_length,
                   char *B2,
                   int B2_length,
                   int reference_length,
                   int *B1_to_reference,
                   int *B2_to_reference,
                   int diagonal_width)
{
  long *d = malloc(sizeof(long) * ((diagonal_width + 2) * 2));
  d[0] = LONG_MAX - 1;
  d[1] = LONG_MAX - 1;
  d[(diagonal_width + 1) * 2] = LONG_MAX - 1;
  d[((diagonal_width + 1) * 2) + 1] = LONG_MAX - 1;
  d += 2;
  for (long i = (diagonal_width / 2); i < diagonal_width; i++) {
    d[i * 2] = i - (diagonal_width / 2);
    d[(i * 2) + 1] = i - (diagonal_width / 2);
  }
  for (long i = 0; i < (diagonal_width / 2); i++) {
    d[i * 2] = (diagonal_width / 2) - i;
    d[(i * 2) + 1] = (diagonal_width / 2) - i;
  }

  int column1 = 1;
  int column2 = 1;
  for (int j = 0; j < reference_length; j++) {
    if (column1 - 1 < B1_length &&
        B1_to_reference[column1 - 1] == j) {
      calculate_column(d, 0, A_length + 1,
                       column1, diagonal_width,
                       A, B1[column1 - 1]);
      column1++;
    }
    if (column2 - 1 < B2_length &&
        B2_to_reference[column2 - 1] == j) {
      calculate_column(d, 1, A_length + 1,
                       column2, diagonal_width,
                       A, B2[column2 - 1]);
      column2++;
    }
    
    if (column1 - 2 < B1_length &&
        column2 - 2 < B2_length &&
        B1_to_reference[column1 - 2] == j &&
        B2_to_reference[column2 - 2] == j) {
      min_column(d, 0, 1, column1 - 1, column2 - 1,
                 A_length + 1, diagonal_width);
      min_column(d, 1, 0, column2 - 1, column1 - 1,
                 A_length + 1, diagonal_width);
    } else {
      if (column1 - 2 < B1_length &&
          column2 - 1 < B2_length &&
          B1_to_reference[column1 - 2] == j &&
          B2_to_reference[column2 - 1] == j + 1) {
        min_column(d, 0, 1, column1 - 1, column2 - 1,
                   A_length + 1, diagonal_width);
      }
      if (column2 - 2 < B2_length &&
          column1 - 1 < B1_length &&
          B2_to_reference[column2 - 2] == j &&
          B1_to_reference[column1 - 1] == j + 1) {
        min_column(d, 1, 0, column2 - 1, column1 - 1,
                   A_length + 1, diagonal_width);
      }
    }

    while (column1 - 1 < B1_length &&
           B1_to_reference[column1 - 1] <= j) {
      calculate_column(d, 0, A_length + 1,
                       column1, diagonal_width,
                       A, B1[column1 - 1]);
      column1++;
    }

    while (column2 - 1 < B2_length &&
           B2_to_reference[column2 - 1] <= j) {
      calculate_column(d, 1, A_length + 1,
                       column2, diagonal_width,
                       A, B2[column2 - 1]);
      column2++;
    }
  }

  long result = min2(d[(A_length - (column1 - 1 - (diagonal_width / 2))) * 2],
                     d[((A_length - (column2 - 1 - (diagonal_width / 2))) * 2) + 1]);
  d -= 2;
  free(d);
  return result;
}

long align(char *A,
           int A_length,
           char *B1,
           int B1_length,
           char *B2,
           int B2_length,
           int reference_length,
           int *B1_to_reference,
           int *B2_to_reference)
{
  int ldiff = max2(abs(A_length - B1_length),
                   abs(A_length - B2_length));
  int x = max2(1, ldiff);
  long k = _align(A, A_length,
                  B1, B1_length,
                  B2, B2_length,
                  reference_length,
                  B1_to_reference,
                  B2_to_reference,
                  (2 * x) + 1);

  while (k > (2 * x) + 1) {
    x = x * 2;
    k = _align(A, A_length,
               B1, B1_length,
               B2, B2_length,
               reference_length,
               B1_to_reference,
               B2_to_reference,
               (2 * x) + 1);
  }

  return k;
}
