#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "util.h"
#include "align.h"

int read_genome(FILE *fp, char **out)
{
  int s = 1024;
  char *o = malloc(sizeof(char) * s);
  
  int i = 0;
  int c = fgetc(fp);
  while (c != EOF) {
    if (c != '\n') {
      if (!(i < s - 1)) {
        s = s * 2;
        o = realloc(o, sizeof(char) * s);
      }
      o[i++] = (char)c;
    }
    c = fgetc(fp);
  }

  o[i] = '\0';
  o = realloc(o, sizeof(char) * (i + 1));
  (*out) = o;
  return i;
}

int read_map(FILE *fp, int **out)
{
  int s = 1024;
  int *o = malloc(sizeof(int) * s);

  int i = 0;
  int e = fscanf(fp, "%d\n", &(o[i]));
  while (e != EOF) {
    o[i] -= 1;
    i++;
    if (!(i < s)) {
      s = s * 2;
      o = realloc(o, sizeof(int) * s);
    }
    e = fscanf(fp, "%d\n", &(o[i]));
  }

  o = realloc(o, sizeof(int) * i);
  (*out) = o;
  return i;
}

FILE *open_or_exit(char *path)
{
  FILE *f = fopen(path, "r");
  if (f == NULL) {
    fprintf(stderr, "could not open file %s\n", path);
    exit(EXIT_FAILURE);
  }
  return f;
}

int main(int argc, char *argv[])
{
  if (argc < 7) {
    fprintf(stderr,
            "Usage: %s diploid_1.txt diploid_2.txt "
            "1toref_map.txt 2toref_map.txt "
            "haploid.txt reference_length\n",
            argv[0]);
    return EXIT_FAILURE;
  }

  FILE *diploid_1_fp = open_or_exit(argv[1]);
  FILE *diploid_2_fp = open_or_exit(argv[2]);
  FILE *map_1_fp = open_or_exit(argv[3]);
  FILE *map_2_fp = open_or_exit(argv[4]);
  FILE *haploid_fp = open_or_exit(argv[5]);

  int reference_length;
  sscanf(argv[6], "%d", &reference_length);
  char *diploid_1;
  char *diploid_2;
  int *map_1;
  int *map_2;
  char *haploid;
  
  int hl = read_genome(haploid_fp, &haploid);
  int d1l = read_genome(diploid_1_fp, &diploid_1);
  int d2l = read_genome(diploid_2_fp, &diploid_2);
  read_map(map_1_fp, &map_1);
  read_map(map_2_fp, &map_2);

  fclose(diploid_1_fp);
  fclose(diploid_2_fp);
  fclose(map_1_fp);
  fclose(map_2_fp);
  fclose(haploid_fp);

  time_t start = time(NULL);
  long k = align(haploid, hl,
                 diploid_1, d1l,
                 diploid_2, d2l,
                 reference_length,
                 map_1, map_2);
  time_t end = time(NULL);
  fprintf(stderr, "time: %f\n", difftime(end, start));
  printf("%ld\n", k);
  
  return EXIT_SUCCESS;
}
