#include <check.h>
#include "util.h"
#include "align.h"
#include "test_align.h"

START_TEST(single_insert_at_start)
{
  int m[] = {1,2};
  ck_assert_int_eq(align("ABC", 3,
                                 "BC", 2,
                                 "BC", 2,
                                 3, m, m), 1);
}
END_TEST

START_TEST(single_insert_at_end)
{
  int m[] = {0,1};
  ck_assert_int_eq(align("ABC", 3,
                                 "AB", 2,
                                 "AB", 2,
                                 3, m, m), 1);
}
END_TEST

START_TEST(single_insert_in_the_middle)
{
  int m[] = {0,2};
  ck_assert_int_eq(align("ABC", 3,
                                 "AC", 2,
                                 "AC", 2,
                                 3, m, m), 1);
}
END_TEST

START_TEST(multiple_inserts_at_start)
{
  int m[] = {3,4,5};
  ck_assert_int_eq(align("ABCDEF", 6,
                                 "DEF", 3,
                                 "DEF", 3,
                                 6, m, m), 3);
}
END_TEST

START_TEST(multiple_inserts_at_end)
{
  int m[] = {0,1,2};
  ck_assert_int_eq(align("ABCDEF", 6,
                                 "ABC", 3,
                                 "ABC", 3,
                                 6, m, m), 3);
}
END_TEST

START_TEST(multiple_inserts_in_the_middle)
{
  int m[] = {0,1,4,5};
  ck_assert_int_eq(align("ABCDEF", 6,
                                 "ABEF", 4,
                                 "ABEF", 4,
                                 6, m, m), 2);
}
END_TEST

START_TEST(multiple_inserts)
{
  int m[] = {1,3,4};
  ck_assert_int_eq(align("ABCDEF", 6,
                                 "BDE", 3,
                                 "BDE", 3,
                                 6, m, m), 3);
}
END_TEST

START_TEST(single_delete_at_start)
{
  int m[] = {0,0,1,2};
  ck_assert_int_eq(align("ABC", 3,
                                 "DABC", 4,
                                 "DABC", 4,
                                 3, m, m), 1);
}
END_TEST

START_TEST(single_delete_at_end)
{
  int m[] = {0,1,2,2};
  ck_assert_int_eq(align("ABC", 3,
                                 "ABCD", 4,
                                 "ABCD", 4,
                                 3, m, m), 1);
}
END_TEST

START_TEST(single_delete_in_the_middle)
{
  int m[] = {0,1,1,2};
  ck_assert_int_eq(align("ABC", 3,
                                 "ABDC", 4,
                                 "ABDC", 4,
                                 3, m, m), 1);
}
END_TEST

START_TEST(multiple_deletes_at_start)
{
  int m[] = {0,0,0,0,1,2};
  ck_assert_int_eq(align("ABC", 3,
                                 "DEFABC", 6,
                                 "DEFABC", 6,
                                 3, m, m), 3);
}
END_TEST

START_TEST(multiple_deletes_at_end)
{
  int m[] = {0,1,2,2,2,2};
  ck_assert_int_eq(align("ABC", 3,
                                 "ABCDEF", 6,
                                 "ABCDEF", 6,
                                 3, m, m), 3);
}
END_TEST

START_TEST(multiple_deletes_in_the_middle)
{
  int m[] = {0,1,1,1,1,2};
  ck_assert_int_eq(align("ABC", 3,
                                 "ABDEFC", 6,
                                 "ABDEFC", 6,
                                 3, m, m), 3);
}
END_TEST

START_TEST(multiple_deletes)
{
  int m[] = {0,0,1,1,1,2};
  ck_assert_int_eq(align("ABC", 3,
                                 "ADBEFC", 6,
                                 "ADBEFC", 6,
                                 3, m, m), 3);
}
END_TEST

START_TEST(single_substitution_at_start)
{
  int m[] = {0,1,2};
  ck_assert_int_eq(align("ABC", 3,
                                 "DBC", 3,
                                 "DBC", 3,
                                 3, m, m), 1);
}
END_TEST

START_TEST(single_substitution_at_end)
{
  int m[] = {0,1,2};
  ck_assert_int_eq(align("ABC", 3,
                                 "ABD", 3,
                                 "ABD", 3,
                                 3, m, m), 1);
}
END_TEST

START_TEST(single_substitution_in_the_middle)
{
  int m[] = {0,1,2};
  ck_assert_int_eq(align("ABC", 3,
                                 "ADC", 3,
                                 "ADC", 3,
                                 3, m, m), 1);
}
END_TEST

START_TEST(multiple_substitutions_at_start)
{
  int m[] = {0,1,2,3,4,5};
  ck_assert_int_eq(align("ABCDEF", 6,
                                 "GHIDEF", 6,
                                 "GHIDEF", 6,
                                 6, m, m), 3);
}
END_TEST

START_TEST(multiple_substitutions_at_end)
{
  int m[] = {0,1,2,3,4,5};
  ck_assert_int_eq(align("ABCDEF", 6,
                                 "ABCGHI", 6,
                                 "ABCGHI", 6,
                                 6, m, m), 3);
}
END_TEST

START_TEST(multiple_substitutions_in_the_middle)
{
  int m[] = {0,1,2,3,4,5};
  ck_assert_int_eq(align("ABCDEF", 6,
                                 "ABGHEF", 6,
                                 "ABGHEF", 6,
                                 6, m, m), 2);
}
END_TEST

START_TEST(multiple_substitutions)
{
  int m[] = {0,1,2,3,4,5};
  ck_assert_int_eq(align("ABCDEF", 6,
                                 "GBHIEF", 6,
                                 "GBHIEF", 6,
                                 6, m, m), 3);
}
END_TEST

START_TEST(single_character_recombination_on_first_character)
{
  int m1[] = {0,1,2};
  int m2[] = {0,1,2};
  ck_assert_int_eq(align("ABB", 3,
                                 "BBB", 3,
                                 "AAA", 3,
                                 3, m1, m2), 0);
  ck_assert_int_eq(align("ABB", 3,
                                 "AAA", 3,
                                 "BBB", 3,
                                 3, m1, m2), 0);
}
END_TEST

START_TEST(single_character_recombination_on_last_character)
{
  int m1[] = {0,1,2};
  int m2[] = {0,1,2};
  ck_assert_int_eq(align("BBB", 3,
                                 "BBA", 3,
                                 "AAB", 3,
                                 3, m1, m2), 0);
  ck_assert_int_eq(align("BBB", 3,
                                 "AAB", 3,
                                 "BBA", 3,
                                 3, m1, m2), 0);
}
END_TEST

START_TEST(single_character_recombination_in_the_middle)
{
  int m1[] = {0,1,2};
  int m2[] = {0,1,2};
  ck_assert_int_eq(align("BBB", 3,
                                 "ABA", 3,
                                 "BAB", 3,
                                 3, m1, m2), 0);
  ck_assert_int_eq(align("BBB", 3,
                                 "BAB", 3,
                                 "ABA", 3,
                                 3, m1, m2), 0);
}
END_TEST

START_TEST(multiple_character_recombination_at_end)
{
  int m1[] = {0,1,2,3,4,5};
  int m2[] = {0,1,2,3,4,5};
  ck_assert_int_eq(align("AAABBB", 6,
                                 "BBBBBB", 6,
                                 "AAAAAA", 6,
                                 6, m1, m2), 0);
  ck_assert_int_eq(align("AAABBB", 6,
                                 "BBBBBB", 6,
                                 "AAAAAA", 6,
                                 6, m1, m2), 0);
}
END_TEST

START_TEST(multiple_character_recombination_at_middle)
{
  int m1[] = {0,1,2,3,4,5};
  int m2[] = {0,1,2,3,4,5};
  ck_assert_int_eq(align("AABBAA", 6,
                                 "BBBBBB", 6,
                                 "AAAAAA", 6,
                                 6, m1, m2), 0);
}
END_TEST

START_TEST(single_character_non_overlapping_recombination_on_first_character)
{
  int m1[] = {1,2};
  int m2[] = {0};
  ck_assert_int_eq(align("ABC", 3,
                                 "BC", 2,
                                 "A", 1,
                                 3, m1, m2), 0);
  ck_assert_int_eq(align("ABC", 3,
                                 "A", 1,
                                 "BC", 2,
                                 3, m2, m1), 0);
}
END_TEST

START_TEST(single_character_non_overlapping_recombination_on_last_character)
{
  int m1[] = {0,1};
  int m2[] = {2};
  ck_assert_int_eq(align("ABC", 3,
                                 "AB", 2,
                                 "C", 1,
                                 3, m1, m2), 0);
  ck_assert_int_eq(align("ABC", 3,
                                 "C", 1,
                                 "AB", 2,
                                 3, m2, m1), 0);
}
END_TEST

START_TEST(single_character_non_overlapping_recombination_in_the_middle)
{
  int m1[] = {0,2};
  int m2[] = {1};
  ck_assert_int_eq(align("ABC", 3,
                                 "AC", 2,
                                 "B", 1,
                                 3, m1, m2), 0);
  ck_assert_int_eq(align("ABC", 3,
                                 "B", 1,
                                 "AC", 2,
                                 3, m2, m1), 0);
}
END_TEST

START_TEST(multiple_character_non_overlapping_recombination_at_end)
{
  int m1[] = {0,1,2};
  int m2[] = {3,4,5};
  ck_assert_int_eq(align("ABCDEF", 6,
                                 "ABC", 3,
                                 "DEF", 3,
                                 6, m1, m2), 0);
  ck_assert_int_eq(align("ABCDEF", 6,
                                 "DEF", 3,
                                 "ABC", 3,
                                 6, m2, m1), 0);
}
END_TEST

START_TEST(multiple_character_non_overlapping_recombination_in_the_middle)
{
  int m1[] = {0,1,4,5};
  int m2[] = {2,3};
  ck_assert_int_eq(align("ABCDEF", 6,
                                 "ABEF", 4,
                                 "CD", 2,
                                 6, m1, m2), 0);
  ck_assert_int_eq(align("ABCDEF", 6,
                                 "CD", 2,
                                 "ABEF", 4,
                                 6, m2, m1), 0);
}
END_TEST

START_TEST(multiple_recombinations)
{
  int m1[] = {1,2,4,5};
  int m2[] = {0,2,3,4};
  ck_assert_int_eq(align("ABCDEF", 6,
                                 "BBEF", 4,
                                 "ACDE", 4,
                                 6, m1, m2), 0);
  ck_assert_int_eq(align("ABCDEF", 6,
                                 "ACDE", 4,
                                 "BBEF", 4,
                                 6, m2, m1), 0);
}
END_TEST

START_TEST(test_only_recombination_1)
{
  int m1[] = {0,1,2,3,4};
  int m2[] = {0,1,2,3,4};

  ck_assert_int_eq(align("ACACA", 5,
                                 "AAAAA", 5,
                                 "CCCCC", 5,
                                 5,
                                 m1, m2), 0);
}
END_TEST

START_TEST(test_only_recombination_2)
{
  int m1[] = {0,1,2,3,3,3,4,5};
  int m2[] = {0,1,1,2,3,4,5};

  ck_assert_int_eq(align("ABCDEF", 6,
                                 "ABCDDDEF", 8,
                                 "ABBCDEF", 7,
                                 6,
                                 m1, m2), 0);
}
END_TEST

START_TEST(test_only_recombination_3)
{
  int m1[] = {0,1,2};
  int m2[] = {2,3};

  ck_assert_int_eq(align("AABB", 4,
                                 "AAB", 3,
                                 "BB", 2,
                                 4,
                                 m1, m2), 0);
}
END_TEST

START_TEST(test_only_recombination_4)
{
  int m1[] = {0,2};
  int m2[] = {0,1};

  ck_assert_int_eq(align("ABA", 3,
                                 "AA", 2,
                                 "BB", 2,
                                 3,
                                 m1, m2), 0);
}
END_TEST

START_TEST(test_only_recombination_5)
{
  int m1[] = {0,2};
  int m2[] = {0,1,2};

  ck_assert_int_eq(align("ABA", 3,
                                 "AA", 2,
                                 "BBB", 3,
                                 3,
                                 m1, m2), 0);
}
END_TEST

START_TEST(test_non_overlapping_recombination_1)
{
  int m1[] = {0,1};
  int m2[] = {2,3};

  ck_assert_int_eq(align("AABB", 4,
                                 "AA", 2,
                                 "BB", 2,
                                 4,
                                 m1, m2), 0);
}
END_TEST

START_TEST(test_non_overlapping_recombination_2)
{
  int m1[] = {0,1,4,5};
  int m2[] = {2,3};

  ck_assert_int_eq(align("AABBCC", 6,
                                 "AACC", 4,
                                 "BB", 2,
                                 6,
                                 m1, m2), 0);
}
END_TEST

START_TEST(test_recombination_and_substitutions)
{
  int m1[] = {0,1,2,3,4};
  int m2[] = {0,1,2,3,4};

  ck_assert_int_eq(align("ACACA", 5,
                                 "AAAAB", 5,
                                 "CCCCC", 5,
                                 5,
                                 m1, m2), 1);
  ck_assert_int_eq(align("ACACA", 5,
                                 "AAAAA", 5,
                                 "ABCCC", 5,
                                 5,
                                 m1, m2), 1);
}
END_TEST

START_TEST(test_recombination_and_inserts)
{
  int m1[] = {0,1,2};
  int m2[] = {0,1,2,3,4};

  ck_assert_int_eq(align("ACACA", 5,
                                 "ACA", 3,
                                 "CCCCC", 5,
                                 5,
                                 m1, m2), 1);

  int mm1[] = {0,1,3,4};
  ck_assert_int_eq(align("ACACA", 5,
                                 "ABBA", 4,
                                 "CCCCC", 5,
                                 5,
                                 mm1, m2), 1);
}
END_TEST

START_TEST(test_recombination_and_deletes)
{
  int m1[] = {0,1,1};
  int m2[] = {0,1};

  ck_assert_int_eq(align("AC", 2,
                                 "ABC", 3,
                                 "CC", 2,
                                 2,
                                 m1, m2), 0);

  int mm1[] = {0,1,1,2,3,4};
  int mm2[] = {0,1,1,2,2};
  ck_assert_int_eq(align("CCACC", 5,
                                 "AABACC", 6,
                                 "CCCAA", 5,
                                 5,
                                 mm1, mm2), 1);
}
END_TEST

START_TEST(test_recombination_with_empty_string)
{
  int m1[] = {0};
  int m2[] = {0,1,2};

  ck_assert_int_eq(align("AAA", 3,
                                 "", 0,
                                 "AAA", 3,
                                 3,
                                 m1, m2), 0);
  ck_assert_int_eq(align("AAA", 3,
                                 "AAA", 3,
                                 "", 0,
                                 3,
                                 m2, m1), 0);
}
END_TEST

Suite *test_align_suite(void)
{
  Suite *s = suite_create("dalign");
  register_test_case("single_insert_at_start", s, single_insert_at_start);
  register_test_case("single_insert_at_end", s, single_insert_at_end);
  register_test_case("single_insert_in_the_middle", s, single_insert_in_the_middle);
  register_test_case("multiple_inserts_at_start", s, multiple_inserts_at_start);
  register_test_case("multiple_inserts_at_end", s, multiple_inserts_at_end);
  register_test_case("multiple_inserts_in_the_middle", s, multiple_inserts_in_the_middle);
  register_test_case("multiple_inserts", s, multiple_inserts);

  register_test_case("single_delete_at_start", s, single_delete_at_start);
  register_test_case("single_delete_at_end", s, single_delete_at_end);
  register_test_case("single_delete_in_the_middle", s, single_delete_in_the_middle);
  register_test_case("multiple_deletes_at_start", s, multiple_deletes_at_start);
  register_test_case("multiple_deletes_at_end", s, multiple_deletes_at_end);
  register_test_case("multiple_deletes_in_the_middle", s, multiple_deletes_in_the_middle);
  register_test_case("multiple_deletes", s, multiple_deletes);

  register_test_case("single_substitution_at_start",
                     s, single_substitution_at_start);
  register_test_case("single_substitution_at_end",
                     s, single_substitution_at_end);
  register_test_case("single_substitution_in_the_middle",
                     s, single_substitution_in_the_middle);
  register_test_case("multiple_substitutions_at_start",
                     s, multiple_substitutions_at_start);
  register_test_case("multiple_substitutions_at_end",
                     s, multiple_substitutions_at_end);
  register_test_case("multiple_substitutions_in_the_middle",
                     s, multiple_substitutions_in_the_middle);
  register_test_case("multiple_substitutions",
                     s, multiple_substitutions);

  register_test_case("single_character_recombination_on_first_character",
                     s, single_character_recombination_on_first_character);
  register_test_case("single_character_recombination_on_last_character",
                     s, single_character_recombination_on_last_character);
  register_test_case("single_character_recombination_in_the_middle",
                     s, single_character_recombination_in_the_middle);
  register_test_case("single_character_non_overlapping_recombination_on_first_character",
                     s, single_character_non_overlapping_recombination_on_first_character);
  register_test_case("single_character_non_overlapping_recombination_on_last_character",
                     s, single_character_non_overlapping_recombination_on_last_character);
  register_test_case("single_character_non_overlapping_recombination_in_the_middle",
                     s, single_character_non_overlapping_recombination_in_the_middle);
  register_test_case("multiple_character_recombination_at_end",
                     s, multiple_character_recombination_at_end);
  register_test_case("multiple_character_recombination_at_middle",
                     s, multiple_character_recombination_at_middle);
  register_test_case("multiple_character_non_overlapping_recombination_at_end",
                     s, multiple_character_non_overlapping_recombination_at_end);
  register_test_case("multiple_character_non_overlapping_recombination_in_the_middle",
                     s, multiple_character_non_overlapping_recombination_in_the_middle);
  register_test_case("multiple_recombinations",
                     s, multiple_recombinations);

  register_test_case("recombination 1", s, test_only_recombination_1);
  register_test_case("recombination 2", s, test_only_recombination_2);
  register_test_case("recombination 3", s, test_only_recombination_3);
  register_test_case("recombination 4", s, test_only_recombination_4);
  register_test_case("recombination 5", s, test_only_recombination_5);
  register_test_case("non overlapping recombination 1",
                     s, test_non_overlapping_recombination_1);
  register_test_case("non overlapping recombination 2",
                     s, test_non_overlapping_recombination_2);
  register_test_case("recombination and substitutions",
                     s, test_recombination_and_substitutions);
  register_test_case("recombination and inserts",
                     s, test_recombination_and_inserts);
  register_test_case("recombination and deletes",
                     s, test_recombination_and_deletes);
  register_test_case("recombination with empty string",
                     s, test_recombination_with_empty_string);
  return s;
}
