#include <stdlib.h>
#include <stdio.h>
#include <check.h>
#include "util.h"
#include "test_generated_haploids.h"
#include "align.h"

int add_errors(char *a,
               int a_length,
               char **b,
               int *b_length,
               char **c,
               int *c_length,
               int **b_mapping,
               int **c_mapping,
               int *errors,
               unsigned int seed)
{
  char *_b = malloc(sizeof(char) * ((a_length * 2) + 1));
  char *_c = malloc(sizeof(char) * ((a_length * 2) + 1));
  int *mb = malloc(sizeof(int) * a_length * 2);
  int *mc = malloc(sizeof(int) * a_length * 2);

  srand(seed);
  int bi = 0;
  int ci = 0;
  for (int i = 0; i < a_length; i++) {
    _b[bi] = a[i];
    mb[bi] = i;
    bi++;
    _c[ci] = a[i];
    mc[ci] = i;
    ci++;

    int both = (rand() / (double)RAND_MAX) > 0.5;
    double r = rand() / (double)RAND_MAX;
    if (r < 0.1) { // insert
      (*errors)++;
      _b[bi] = a[i];
      mb[bi] = i;
      bi++;
      if (both) {
        _c[ci] = a[i];
        mc[ci] = i;
        ci++;
      }
    } else if (r < 0.2) { // substitution
      (*errors)++;
      bi--;
      _b[bi] = (a[i] == 'A' ? 'T' : 'A');
      mb[bi] = i;
      bi++;
      if (both) {
        ci--;
        _c[ci] = (a[i] == 'A' ? 'T' : 'A');
        mc[ci] = i;
        ci++;
      }
    } else if (r < 0.3) { // delete
      (*errors)++;
      bi--;
      if (both) {
        ci--;
      }
    }
  }
  _b[bi] = '\0';
  _c[ci] = '\0';

  *b = malloc(sizeof(char) * (bi + 1));
  *c = malloc(sizeof(char) * (ci + 1));
  *b_length = bi;
  *c_length = ci;
  *b_mapping = malloc(sizeof(int) * bi);
  *c_mapping = malloc(sizeof(int) * ci);

  memcpy(*b, _b, sizeof(char) * (bi + 1));
  memcpy(*c, _c, sizeof(char) * (ci + 1));
  memcpy(*b_mapping, mb, sizeof(int) * bi);
  memcpy(*c_mapping, mc, sizeof(int) * ci);
  free(_b);
  free(_c);
  free(mb);
  free(mc);

  return 0;
}

START_TEST(test_generated_haploid)
{
  char d[] = "ATCGATCGTCAGCTTAGCTAGCTG";
  int n = sizeof(d) - 1;
  char *b1;
  char *b2;
  int l1;
  int l2;
  int *m1;
  int *m2;
  
  for (int i = 0; i < 100; i++) {
    int e = 0;
    add_errors(d, n,
               &b1, &l1,
               &b2, &l2,
               &m1, &m2,
               &e, i);
    
    ck_assert_int_le(align(d, n,
                           b1, l1,
                           b2, l2,
                           n, m1, m2),
                     e);

    free(b1);
    free(b2);
    free(m1);
    free(m2);
  }
}
END_TEST

Suite *test_generated_haploids_suite(void)
{
  Suite *s = suite_create("generated haploids");
  register_test_case("generated haploid",
                     s, test_generated_haploid);
  return s;
}
